# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#   All Rights Reserved
#
# License:  BSD-3-Clause
"""SRON Python package `pyspex`.

It contains software to read PACE HKT products and SPEXone Level-0 products,
and read/write SPEXone Level-1A products.
"""
